# Copyright (c) 2017 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from setuptools import Extension, setup
import numpy
from os import name as os_name

setup(
    ext_modules=[
        Extension(
            "_nnef",
            sources=["nnef/nnef.cpp", "nnef/cpp/src/nnef.cpp"],
            include_dirs=["nnef/cpp/include", numpy.get_include()],
            language="c++",
            extra_compile_args=["-std=c++11"] if os_name != "nt" else [],
        )
    ],
)
